services.factory('AppServiceManager', function ($q, $cacheFactory, $injector){
	
	return ({
		getData : getData,
		deleteFromCache : deleteFromCache
	});
	
	// Public Methods
	
	/**
	 * Hace llamadas a métodos de otros servicios que devuelven información, 
	 * guardando en cache por defecto. Podemos deshabilitar la cache con el
	 * parámetro cache enabled
	 * 
	 * @param {string} serviceName
	 * @param {string} serviceMethod
	 * @param {Array} arrayParams
	 * @param {boolean} cacheEnabled
	 * @returns {$q@call;defer.promise}
	 */
	function getData(serviceName, serviceMethod, arrayParams, cacheEnabled){
		
		var begin_time = new Date().getTime();
		
		cacheEnabled = (typeof cacheEnabled  === 'undefined') ? true : cacheEnabled;
		
		var keyCache = getCacheKey(serviceName, serviceMethod, arrayParams);

		if(cacheEnabled){
			
			result = getDataFromCache(keyCache);

			if(result){
				var deferred = $q.defer();
				deferred.resolve(result);
				return deferred.promise;
			}
		}

		promise = _getData(serviceName, serviceMethod, arrayParams);
		promise.success(function(response){

				console.log(
					'ElapsedTime of ' +
					serviceName + ' ' +
					serviceMethod + ' ' +
					(new Date().getTime() - begin_time) + 
					' ms'
				);
				if(cacheEnabled){
					saveDataToCache(keyCache, response);
				}

				return response;
		});
		
		
		return ( promise.then(handleSuccess, handleError) );
	}
	
	function deleteFromCache(serviceName, serviceMethod, arrayParams){
		
		var keyCache = getCacheKey(serviceName, serviceMethod, arrayParams);
		
		_deleteFromCache(keyCache);
	}
	
	// Private methods
	
	/**
	 * Trata una respuesta a la API remota devolviendo el objeto JSON que nos
	 * interesa
	 * @param {object} response
	 * @returns {objet} JSON data
	 */
	function handleSuccess(response){
		return response.data;
	}
	
	/**
	 * Tratamiento de una respuesta erronea
	 * TODO: redirigir a un controllador de error, crear en la API un apartado
	 * para hacer un log de las respuestas de error de los dispositivos. De esta
	 * manera podremos hacer un tratamiento de errores en distintos dispositivos.
	 * 
	 * @param {object} response
	 * @returns {undefined}
	 */
	function handleError(response){
		// Red del telefono no disponible
		if(response.status === 0){
			document.dispatchEvent(eventNetworkProblem);
		}
	}
	
	/**
	 * Hace la petición a un servicio serviceName->serviceMethod(param1,param2,..) y
	 * devuelve la respuesta, en este caso objectos promise de AngularJS
	 * 
	 * @param {string} serviceName
	 * @param {string} serviceMethod
	 * @param {array} arrayParams
	 * @returns {object} Promise
	 */
	function _getData(serviceName, serviceMethod, arrayParams){
		
		service = $injector.get(serviceName);
		
		return service[serviceMethod].apply(service, arrayParams);
	}
	
	/**
	 * Devuelve la clave única para poder cachear las respuestas en la aplicación
	 * 
	 * @param {type} serviceName
	 * @param {type} serviceMethod
	 * @param {type} arrayParams
	 * @returns {String}
	 */
	function getCacheKey(serviceName, serviceMethod, arrayParams){
		
		return serviceName + serviceMethod + JSON.stringify(arrayParams);
	}
	
	/**
	 * Intenta conseguir la información de la cache
	 * @param {string} key
	 * @returns {res|Boolean}
	 */
	function getDataFromCache(key){
		
		CustomCache = getCacheObj();
		res = CustomCache.get(key);

		return (typeof res !== 'undefined') ? res : false;
	}
	
	/**
	 * @param {string} key
	 * @param {mixed} data
	 * @returns {undefined}
	 */
	function saveDataToCache(key, data){
		
		CustomCache =  getCacheObj();
		CustomCache.put(key, data);
	}
	
	function _deleteFromCache(key){
		CustomCache =  getCacheObj();
		CustomCache.remove(key);
	}
	
	/**
	 * @returns {CustomCache} Pseudo Singleton 
	 */
	function getCacheObj(){
		
		CustomCache =  $cacheFactory.get('ApiCache');

		if(typeof CustomCache === 'undefined'){
			
			CustomCache = $cacheFactory('ApiCache');
		}

		return CustomCache;
	}
});